from __future__ import absolute_import


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 15:26:16 GMT-0500$"


import itertools
import operator
import warnings

import kenjutsu.format


def split_blocks(space_shape, block_shape, block_halo=None):
    """
        Return a list of slicings to cut each block out of an array or other.

        Takes an array with ``space_shape`` and ``block_shape`` for every
        dimension and a ``block_halo`` to extend each block on each side. From
        this, it can compute slicings to use for cutting each block out from
        the original array, HDF5 dataset or other.

        Note:
            Blocks on the boundary that cannot extend the full range will
            be truncated to the largest block that will fit. This will raise
            a warning, which can be converted to an exception, if needed.

        Args:
            space_shape(tuple):            Shape of array to slice
            block_shape(tuple):            Size of each block to take
            block_halo(tuple):             Halo to tack on to each block

        Returns:
            collections.Sequence of \
            tuples of slices:              Provides tuples of slices for \
                                           retrieving blocks.

        Examples:

            >>> split_blocks(
            ...     (2, 3,), (1, 1,), (1, 1,)
            ... )  #doctest: +NORMALIZE_WHITESPACE
            ([(slice(0, 1, 1), slice(0, 1, 1)),
              (slice(0, 1, 1), slice(1, 2, 1)),
              (slice(0, 1, 1), slice(2, 3, 1)),
              (slice(1, 2, 1), slice(0, 1, 1)),
              (slice(1, 2, 1), slice(1, 2, 1)),
              (slice(1, 2, 1), slice(2, 3, 1))],
            <BLANKLINE>
             [(slice(0, 2, 1), slice(0, 2, 1)),
              (slice(0, 2, 1), slice(0, 3, 1)),
              (slice(0, 2, 1), slice(1, 3, 1)),
              (slice(0, 2, 1), slice(0, 2, 1)),
              (slice(0, 2, 1), slice(0, 3, 1)),
              (slice(0, 2, 1), slice(1, 3, 1))],
            <BLANKLINE>
             [(slice(0, 1, 1), slice(0, 1, 1)),
              (slice(0, 1, 1), slice(1, 2, 1)),
              (slice(0, 1, 1), slice(1, 2, 1)),
              (slice(1, 2, 1), slice(0, 1, 1)),
              (slice(1, 2, 1), slice(1, 2, 1)),
              (slice(1, 2, 1), slice(1, 2, 1))])

    """

    try:
        irange = xrange
    except NameError:
        irange = range

    try:
        from itertools import ifilter, imap
    except ImportError:
        ifilter, imap = filter, map

    if block_halo is not None:
        if not (len(space_shape) == len(block_shape) == len(block_halo)):
            raise ValueError(
                "The dimensions of `space_shape`, `block_shape`, and"
                " `block_halo` should be the same."
            )
    else:
        if not (len(space_shape) == len(block_shape)):
            raise ValueError(
               "The dimensions of `space_shape` and `block_shape` should be"
               " the same."
            )

        block_halo = tuple()
        for i in irange(len(space_shape)):
            block_halo += (0,)

    vec_add = lambda a, b: imap(operator.add, a, b)
    vec_sub = lambda a, b: imap(operator.sub, a, b)

    vec_mul = lambda a, b: imap(operator.mul, a, b)
    vec_mod = lambda a, b: imap(operator.mod, a, b)

    vec_nonzero = lambda a: \
            imap(lambda _: _[0], ifilter(lambda _: _[1], enumerate(a)))
    vec_str = lambda a: imap(str, a)

    vec_clip_floor = lambda a, a_min: \
            imap(lambda _: _ if _ >= a_min else a_min, a)
    vec_clip_ceil = lambda a, a_max: \
            imap(lambda _: _ if _ <= a_max else a_max, a)
    vec_clip = lambda a, a_min, a_max: \
            vec_clip_ceil(vec_clip_floor(a, a_min), a_max)

    uneven_block_division = tuple(vec_mod(space_shape, block_shape))

    if any(uneven_block_division):
        uneven_block_division_str = vec_nonzero(uneven_block_division)
        uneven_block_division_str = vec_str(uneven_block_division_str)
        uneven_block_division_str = ", ".join(uneven_block_division_str)

        warnings.warn(
            "Blocks will not evenly divide the array." +
            " The following dimensions will be unevenly divided: %s." %
            uneven_block_division_str,
            RuntimeWarning
        )

    ranges_per_dim = []
    haloed_ranges_per_dim = []
    trimmed_halos_per_dim = []

    for each_dim in irange(len(space_shape)):
        # Construct each block using the block size given. Allow to spill over.
        if block_shape[each_dim] == -1:
            block_shape = (block_shape[:each_dim] +
                           space_shape[each_dim:each_dim+1] +
                           block_shape[each_dim+1:])

        # Generate block ranges.
        a_range = []
        for i in irange(2):
            offset = i * block_shape[each_dim]
            this_range = irange(
                offset,
                offset + space_shape[each_dim],
                block_shape[each_dim]
            )
            a_range.append(list(this_range))

        # Add the halo to each block on both sides.
        a_range_haloed = []
        for i in irange(2):
            sign = 2 * i - 1

            haloed = vec_mul(
                itertools.repeat(sign, len(a_range[i])),
                itertools.repeat(block_halo[each_dim], len(a_range[i])),
            )
            haloed = vec_add(a_range[i], haloed)
            haloed = vec_clip(haloed, 0, space_shape[each_dim])

            a_range_haloed.append(list(haloed))

        # Compute how to trim the halo off of each block.
        # Clip each block to the boundaries.
        a_trimmed_halo = []
        for i in irange(2):
            trimmed = vec_sub(a_range[i], a_range_haloed[0])
            a_trimmed_halo.append(list(trimmed))
            a_range[i] = list(vec_clip(a_range[i], 0, space_shape[each_dim]))

        # Convert all ranges to slices for easier use.
        a_range = tuple(imap(slice, *a_range))
        a_range_haloed = tuple(imap(slice, *a_range_haloed))
        a_trimmed_halo = tuple(imap(slice, *a_trimmed_halo))

        # Format all slices.
        a_range = kenjutsu.format.reformat_slices(a_range)
        a_range_haloed = kenjutsu.format.reformat_slices(a_range_haloed)
        a_trimmed_halo = kenjutsu.format.reformat_slices(a_trimmed_halo)

        # Collect all blocks
        ranges_per_dim.append(a_range)
        haloed_ranges_per_dim.append(a_range_haloed)
        trimmed_halos_per_dim.append(a_trimmed_halo)

    # Take all combinations of all ranges to get blocks.
    orig_blocks = list(itertools.product(*ranges_per_dim))
    haloed_blocks = list(itertools.product(*haloed_ranges_per_dim))
    trimmed_halos = list(itertools.product(*trimmed_halos_per_dim))

    return(orig_blocks, haloed_blocks, trimmed_halos)



from __future__ import absolute_import

__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 14:49:29 GMT-0500$"


import numbers
import math

import kenjutsu.format


class UnknownSliceLengthException(Exception):
    """
        Raised if a slice does not have a known length.
    """

    pass


def len_slice(a_slice, a_length=None):
    """
        Determines how many elements a slice will contain.

        Raises:
            UnknownSliceLengthException: Will raise an exception if
            a_slice.stop and a_length is None.

        Args:
            a_slice(slice):        a slice to reformat.

            a_length(int):         a length to fill for stopping if not
                                   provided.

        Returns:
            (slice):               a new slice with as many values filled in as
                                   possible.

        Examples:

            >>> len_slice(slice(2, None), 10)
            8

            >>> len_slice(slice(2, 6))
            4
    """

    if isinstance(a_slice, numbers.Integral):
        raise TypeError(
            "An integral index does not provide an object with a length."
        )

    new_slice = kenjutsu.format.reformat_slice(a_slice, a_length)

    new_slice_size = 0
    if isinstance(new_slice, slice):
        if (new_slice.step > 0 and new_slice.start >= 0 and
                (new_slice.stop is None or new_slice.stop < 0)):
                raise UnknownSliceLengthException(
                    "Cannot determine slice length without a defined end"
                    " point. The reformatted slice was %s." % repr(new_slice)
                )
        elif (new_slice.step < 0 and new_slice.start < 0 and
                (new_slice.stop is None or new_slice.stop >= 0)):
                raise UnknownSliceLengthException(
                    "Cannot determine slice length without a defined start"
                    " point. The reformatted slice was %s." % repr(new_slice)
                )

        if new_slice.step < 0 and new_slice.stop is None:
            new_slice = slice(new_slice.start, -1, new_slice.step)

        new_slice_diff = float(new_slice.stop - new_slice.start)

        new_slice_size = int(math.ceil(new_slice_diff / new_slice.step))
    else:
        new_slice_size = len(new_slice)

    return(new_slice_size)


def len_slices(slices, lengths=None):
    """
        Takes a tuple of slices and reformats them to fill in as many undefined
        values as possible.

        Args:
            slices(tuple(slice)):        a tuple of slices to reformat.
            lengths(tuple(int)):         a tuple of lengths to fill.

        Returns:
            (slice):                     a tuple of slices with all default
                                         values filled if possible.

        Examples:

            >>> len_slices(
            ...     (
            ...         slice(None),
            ...         slice(3, None),
            ...         slice(None, 5),
            ...         slice(None, None, 2)
            ...     ),
            ...     (10, 13, 15, 20)
            ... )
            (10, 10, 5, 10)
    """

    new_slices = kenjutsu.format.reformat_slices(slices, lengths)

    lens = []

    for each_slice in new_slices:
        if not isinstance(each_slice, numbers.Integral):
            lens.append(len_slice(each_slice))

    lens = tuple(lens)

    return(lens)



#!/usr/bin/env python

# -*- coding: utf-8 -*-

__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 14:49:29 GMT-0500$"


import doctest
import itertools
import operator
import sys
import unittest

from kenjutsu import measure


try:
    irange = xrange
except NameError:
    irange = range


# Load doctests from `measure`.
def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(measure))
    return tests


class TestMeasure(unittest.TestCase):
    def setUp(self):
        pass


    def test_len_slice(self):
        with self.assertRaises(measure.UnknownSliceLengthException):
            measure.len_slice(slice(None))

        with self.assertRaises(measure.UnknownSliceLengthException):
            measure.len_slice(slice(0, -1, 1))

        with self.assertRaises(measure.UnknownSliceLengthException):
            measure.len_slice(slice(None, None, -1))

        with self.assertRaises(measure.UnknownSliceLengthException):
            measure.len_slice(slice(-1, 1, -1))

        for size in [10, 11, 12]:
            excess = size + 3
            each_range = range(size)
            for start in itertools.chain([None], irange(-excess, excess)):
                for stop in itertools.chain([None], irange(-excess, excess)):
                    for step in itertools.chain(irange(-excess, excess)):
                        step = None if step == 0 else step

                        a_slice = slice(start, stop, step)

                        l = measure.len_slice(a_slice, size)
                        self.assertEqual(
                            l,
                            len(each_range[a_slice])
                        )

                        a_slice = list()
                        a_slice.append(0 if start is None else start)
                        a_slice.append(0 if stop is None else stop)
                        a_slice.append(0 if step is None else step)

                        a_op = operator.itemgetter(*a_slice)

                        expected_result = None
                        try:
                            expected_result = a_op(each_range)
                        except IndexError:
                            pass

                        if expected_result is not None:
                            l = measure.len_slice(a_slice, size)
                            self.assertEqual(len(expected_result), l)

                if start is not None:
                    a_slice = start

                    with self.assertRaises(TypeError):
                        measure.len_slice(a_slice, size)

            self.assertEqual(
                measure.len_slice(Ellipsis, size),
                len(each_range[:])
            )

            self.assertEqual(
                measure.len_slice(tuple(), size),
                len(each_range[:])
            )


    def test_len_slices(self):
        with self.assertRaises(measure.UnknownSliceLengthException):
            measure.len_slices((
                slice(None),
                slice(3, None),
                slice(None, 5),
                slice(None, None, 2)
            ))

        l = measure.len_slices(Ellipsis, 10)
        self.assertEqual(
            l,
            (10,)
        )

        l = measure.len_slices(tuple(), 10)
        self.assertEqual(
            l,
            (10,)
        )

        l = measure.len_slices(slice(None), 10)
        self.assertEqual(
            l,
            (10,)
        )

        l = measure.len_slices((slice(None),), 10)
        self.assertEqual(
            l,
            (10,)
        )

        l = measure.len_slices(
            (
                -1,
                slice(None),
                slice(3, None),
                slice(None, 5),
                slice(None, None, 2),
                [-1, -2, -1, 1, 5]
            ),
            (12, 10, 13, 15, 20, 10)
        )
        self.assertEqual(
            l,
            (10, 10, 5, 10, 5)
        )

        l = measure.len_slices(
            Ellipsis,
            (2, 3, 4, 5)
        )
        self.assertEqual(
            l,
            (2, 3, 4, 5)
        )

        l = measure.len_slices(
            (
                Ellipsis,
                slice(0, 1)
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            l,
            (2, 3, 4, 1)
        )

        l = measure.len_slices(
            (
                slice(0, 1),
                Ellipsis
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            l,
            (1, 3, 4, 5)
        )

        l = measure.len_slices(
            (
                slice(0, 1),
                Ellipsis,
                slice(0, 1)
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            l,
            (1, 3, 4, 1)
        )

        l = measure.len_slices(
            (
                slice(0, 1),
                Ellipsis,
                slice(0, 1),
                slice(0, 1),
                slice(0, 1)
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            l,
            (1, 1, 1, 1)
        )


    def tearDown(self):
        pass



if __name__ == '__main__':
    sys.exit(unittest.main())



#!/usr/bin/env python

# -*- coding: utf-8 -*-

__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 11:35:58 GMT-0500$"


import doctest
import sys
import unittest

from kenjutsu import core


try:
    irange = xrange
except NameError:
    irange = range


# Load doctests from `core`.
def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(core))
    return tests


class TestCore(unittest.TestCase):
    def setUp(self):
        pass


    def tearDown(self):
        pass



if __name__ == '__main__':
    sys.exit(unittest.main())



__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 14:20:52 GMT-0500$"


import collections
import itertools
import numbers


def index_to_slice(index):
    """
        Convert an index to a slice.

        Note:
            A single index behaves differently from a length 1 ``slice``. When
            applying the former one reduces that dimension; whereas, applying
            the latter results in a singleton dimension being retained. Also
            if an index is out of bounds, one gets an ``IndexError``. However,
            with an out of bounds length 1 ``slice``, one simply doesn't get
            the requested range.

        Args:
            index(int):                  an index to convert to a slice

        Returns:
            (slice):                     a slice corresponding to the index

        Examples:

            >>> index_to_slice(1)
            slice(1, 2, 1)

            >>> index_to_slice(-1)
            slice(-1, -2, -1)
    """

    if not isinstance(index, numbers.Integral):
        raise TypeError(
            "Expected an integral type. Instead got `%s`." % str(index)
        )

    step = -1 if index < 0 else 1

    return slice(index, index + step, step)


def reformat_slice(a_slice, a_length=None):
    """
        Takes a slice and reformats it to fill in as many undefined values as
        possible.

        Args:
            a_slice(slice):        a slice to reformat.

            a_length(int):         a length to fill for stopping if not
                                   provided.

        Returns:
            (slice):               a new slice with as many values filled in as
                                   possible.

        Examples:

            >>> reformat_slice(slice(2, -1, None))
            slice(2, -1, 1)

            >>> reformat_slice(slice(2, -1, None), 10)
            slice(2, 9, 1)
    """

    new_slice = a_slice
    if (new_slice is Ellipsis) or (new_slice == tuple()):
        new_slice = slice(None)
    elif isinstance(a_slice, numbers.Integral):
        new_slice = index_to_slice(a_slice)
    elif isinstance(a_slice, collections.Sequence):
        if not all(map(lambda i: isinstance(i, numbers.Integral), a_slice)):
            raise ValueError(
                "Arbitrary sequences not permitted."
                " All elements must be of integral type."
            )

        # Normalize each integer in the range.
        new_slice = []
        for i in a_slice:
            new_slice.append(reformat_slice(i, a_length))
        return new_slice
    elif not isinstance(a_slice, slice):
        raise ValueError(
            "Expected a `slice` type. Instead got `%s`." % str(a_slice)
        )

    if new_slice.step == 0:
        raise ValueError("Slice cannot have a step size of `0`.")

    start = new_slice.start
    stop = new_slice.stop
    step = new_slice.step

    # Fill unknown values.
    if step is None:
        step = 1
    if start is None:
        if step > 0:
            start = 0
        elif step < 0:
            start = -1
    if (stop is None) and (step > 0):
        stop = a_length

    stop_i = stop is not None

    # Make adjustments for length
    if a_length is not None:
        # Normalize out-of-bound step sizes.
        if step < -a_length:
            step = -a_length
        elif step > a_length:
            step = a_length

        # Normalize bounded negative values.
        if -a_length <= start < 0:
            start += a_length
        if stop_i and (-a_length <= stop < 0):
            stop += a_length

        # Handle out-of-bound limits.
        if step > 0:
            if (start > a_length) or (stop < -a_length):
                start = stop = 0
                step = 1
            else:
                if start < -a_length:
                    start = 0
                if stop > a_length:
                    stop = a_length
        elif step < 0:
            if (start < -a_length) or (stop_i and stop >= (a_length - 1)):
                start = stop = 0
                step = 1
            else:
                if start >= a_length:
                    start = a_length - 1
                if stop_i and stop < -a_length:
                    stop = None
                    stop_i = False

    # Catch some known empty slices.
    if stop_i and (start == stop):
        start = stop = 0
        step = 1
    elif (step > 0) and (stop == 0):
        start = stop = 0
        step = 1
    elif (step < 0) and (stop == -1):
        start = stop = 0
        step = 1
    elif stop_i and (start >= 0) and (stop >= 0):
        if (step > 0) and (start > stop):
            start = stop = 0
            step = 1
        elif (step < 0) and (start < stop):
            start = stop = 0
            step = 1

    new_slice = slice(start, stop, step)
    if isinstance(a_slice, numbers.Integral):
        if new_slice.start == new_slice.stop == 0:
            raise IndexError("Index out of range.")
        new_slice = new_slice.start

    return(new_slice)


def reformat_slices(slices, lengths=None):
    """
        Takes a tuple of slices and reformats them to fill in as many undefined
        values as possible.

        Args:
            slices(tuple(slice)):        a tuple of slices to reformat.
            lengths(tuple(int)):         a tuple of lengths to fill.

        Returns:
            (slice):                     a tuple of slices with all default
                                         values filled if possible.

        Examples:

            >>> reformat_slices(
            ...     (
            ...         slice(None),
            ...         slice(3, None),
            ...         slice(None, 5),
            ...         slice(None, None, 2)
            ...     ),
            ...     (10, 13, 15, 20)
            ... )
            (slice(0, 10, 1), slice(3, 13, 1), slice(0, 5, 1), slice(0, 20, 2))
    """

    new_slices = slices
    if new_slices == tuple():
        new_slices = Ellipsis

    try:
        len(new_slices)
    except TypeError:
        new_slices = (new_slices,)

    new_lengths = lengths
    try:
        if new_lengths is not None:
            len(new_lengths)
    except TypeError:
        new_lengths = (new_lengths,)

    el_idx = None
    try:
        el_idx = new_slices.index(Ellipsis)
    except ValueError:
        pass

    if new_lengths is not None and el_idx is None:
        if len(new_slices) != len(new_lengths):
            raise ValueError("Shape must be the same as the number of slices.")
    elif new_lengths is not None:
        if (len(new_slices) - 1) > len(new_lengths):
            raise ValueError(
                "Shape must be as large or larger than the number of slices"
                " without the Ellipsis."
            )

    if el_idx is not None:
        # Break into three cases.
        #
        # 1. Before the Ellipsis
        # 2. The Ellipsis
        # 3. After the Ellipsis
        #
        # Cases 1 and 3 are trivially solved as before.
        # Case 2 is either a no-op or a bunch of `slice(None)`s.
        #
        # The result is a combination of all of these.

        slices_before = new_slices[:el_idx]
        slices_after = new_slices[el_idx+1:]

        if Ellipsis in slices_before or Ellipsis in slices_after:
            raise ValueError("Only one Ellipsis is permitted. Found multiple.")

        new_lengths_before = None
        new_lengths_after = None
        slice_el = (Ellipsis,)
        if new_lengths is not None:
            pos_before = len(slices_before)
            pos_after = len(new_lengths) - len(slices_after)

            new_lengths_before = new_lengths[:pos_before]
            new_lengths_after = new_lengths[pos_after:]

            new_lengths_el = new_lengths[pos_before:pos_after]
            slice_el = len(new_lengths_el) * (slice(None),)
            if slice_el:
                slice_el = reformat_slices(
                    slice_el,
                    new_lengths_el
                )

        if slices_before:
            slices_before = reformat_slices(slices_before, new_lengths_before)
        if slices_after:
            slices_after = reformat_slices(slices_after, new_lengths_after)

        new_slices = slices_before + slice_el + slices_after
    else:
        if new_lengths is None:
            new_lengths = [None] * len(new_slices)

        new_slices = list(new_slices)
        for i, each_length in enumerate(new_lengths):
            new_slices[i] = reformat_slice(new_slices[i], each_length)

        new_slices = tuple(new_slices)

    n_seqs = sum(map(
        lambda i: isinstance(i, collections.Sequence), new_slices
    ))
    if n_seqs > 1:
        raise ValueError(
            "Only one integral sequence supported."
            " Instead got `%s`." % str(n_seqs)
        )

    return(new_slices)


def split_indices(slices):
    """
        Splits slices with multiple indices into multiple splices.

        Support of slices with a sequence of indices is varied. Some
        libraries like NumPy support them without issues. Other libraries
        like h5py support them as long as they are in sequential order.
        In still other libraries support is non-existent. However, in
        all those cases normally a single index is permissible. This
        converts slices with multiple indices into a list of slices with
        a single index each. While this still leaves it up to the user
        to iterate over these and combine the results in some sensible
        way, it is better than just getting a failure and should extend
        well to a variety of cases.

        Args:
            slices(tuple(slice)):        a tuple of slices to split

        Returns:
            (list(tuple(slice))):        a list of a tuple of slices

        Examples:

            >>> split_indices(
            ...     (
            ...         3,
            ...         Ellipsis,
            ...         [0, 1, 2],
            ...         slice(2, 5),
            ...         slice(4, 6, 2)
            ...     )
            ... )  # doctest: +NORMALIZE_WHITESPACE
            [(3, Ellipsis, slice(0, 1, 1), slice(2, 5, 1), slice(4, 6, 2)),
             (3, Ellipsis, slice(1, 2, 1), slice(2, 5, 1), slice(4, 6, 2)),
             (3, Ellipsis, slice(2, 3, 1), slice(2, 5, 1), slice(4, 6, 2))]
    """

    ref_slices = reformat_slices(slices)

    mtx_slices = []
    for each_dim_slice in ref_slices:
        if each_dim_slice is Ellipsis:
            mtx_slices.append([each_dim_slice])
        elif isinstance(each_dim_slice, numbers.Integral):
            mtx_slices.append([each_dim_slice])
        elif isinstance(each_dim_slice, slice):
            mtx_slices.append([each_dim_slice])
        elif isinstance(each_dim_slice, collections.Sequence):
            new_slice = []
            for i in each_dim_slice:
                new_slice.append(index_to_slice(i))
            mtx_slices.append(new_slice)

    result_slices = []
    for each_dim_slice in itertools.product(*mtx_slices):
        result_slices.append(tuple(each_dim_slice))

    return result_slices



# -*- coding: utf-8 -*-

"""
The module ``core`` provides support for working with ``slice``\ s.

===============================================================================
Overview
===============================================================================
The module ``core`` provides several functions that are useful for working
with a Python ``slice`` or ``tuple`` of ``slice``\ s. This is of particular
value when working with NumPy_.

.. _NumPy: http://www.numpy.org/

===============================================================================
API
===============================================================================
"""


from __future__ import absolute_import


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 11:35:58 GMT-0500$"


import kenjutsu.blocks
import kenjutsu.format
import kenjutsu.measure


reformat_slice = kenjutsu.format.reformat_slice
reformat_slices = kenjutsu.format.reformat_slices

UnknownSliceLengthException = kenjutsu.measure.UnknownSliceLengthException
len_slice = kenjutsu.measure.len_slice
len_slices = kenjutsu.measure.len_slices

split_blocks = kenjutsu.blocks.split_blocks



__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 08, 2016 14:20:52 GMT-0500$"


import doctest
import itertools
import math
import operator
import unittest

from kenjutsu import format


try:
    irange = xrange
except NameError:
    irange = range


# Load doctests from `format`.
def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(format))
    return tests


class TestFormat(unittest.TestCase):
    def setUp(self):
        pass


    def test_index_to_slice(self):
        with self.assertRaises(TypeError) as e:
            format.index_to_slice(None)

        self.assertEqual(
            str(e.exception),
            "Expected an integral type. Instead got `None`."
        )

        with self.assertRaises(TypeError) as e:
            format.index_to_slice(2.5)

        self.assertEqual(
            str(e.exception),
            "Expected an integral type. Instead got `2.5`."
        )

        with self.assertRaises(TypeError) as e:
            format.index_to_slice((0,))

        self.assertEqual(
            str(e.exception),
            "Expected an integral type. Instead got `(0,)`."
        )

        with self.assertRaises(TypeError) as e:
            format.index_to_slice([0, 1])

        self.assertEqual(
            str(e.exception),
            "Expected an integral type. Instead got `[0, 1]`."
        )

        with self.assertRaises(TypeError) as e:
            format.index_to_slice(slice(None))

        self.assertEqual(
            str(e.exception),
            "Expected an integral type. Instead got `slice(None, None, None)`."
        )

        with self.assertRaises(TypeError) as e:
            format.index_to_slice(Ellipsis)

        self.assertEqual(
            str(e.exception),
            "Expected an integral type. Instead got `Ellipsis`."
        )

        for size in [10, 11, 12]:
            excess = size + 3
            each_range = range(size)

            for index in itertools.chain(irange(-excess, excess)):
                expected_result = []
                try:
                    expected_result = [each_range[index]]
                except IndexError:
                    pass

                rf_slice = format.index_to_slice(index)

                self.assertIsInstance(rf_slice, slice)

                result = list(each_range[rf_slice])
                self.assertEqual(result, expected_result)

                start = rf_slice.start
                stop = rf_slice.stop
                step = rf_slice.step

                self.assertEqual(int(math.copysign(1, index)), step)

                l = float(stop - start)/float(step)
                self.assertEqual(
                    int(math.ceil(l)),
                    1
                )


    def test_reformat_slice(self):
        with self.assertRaises(ValueError) as e:
            format.reformat_slice(None)

        self.assertEqual(
            str(e.exception),
            "Expected a `slice` type. Instead got `None`."
        )

        with self.assertRaises(ValueError) as e:
            format.reformat_slice(slice(None, None, 0))

        self.assertEqual(
            str(e.exception),
            "Slice cannot have a step size of `0`."
        )

        with self.assertRaises(ValueError) as e:
            format.reformat_slice([None])

        self.assertEqual(
            str(e.exception),
            "Arbitrary sequences not permitted."
            " All elements must be of integral type."
        )

        for size in [10, 11, 12]:
            excess = size + 3
            each_range = range(size)

            for start in itertools.chain([None], irange(-excess, excess)):
                for stop in itertools.chain([None], irange(-excess, excess)):
                    for step in itertools.chain(irange(-excess, excess)):
                        step = None if step == 0 else step

                        a_slice = slice(start, stop, step)

                        rf_slice = format.reformat_slice(a_slice)
                        self.assertEqual(
                            each_range[a_slice],
                            each_range[rf_slice]
                        )

                        rf_slice = format.reformat_slice(a_slice, size)
                        self.assertEqual(
                            each_range[a_slice],
                            each_range[rf_slice]
                        )

                        new_start = rf_slice.start
                        new_stop = rf_slice.stop
                        new_step = rf_slice.step

                        if (new_step is not None and
                            new_step < 0 and
                            new_stop is None):
                            new_stop = -1

                        l = float(new_stop - new_start)/float(new_step)
                        self.assertEqual(
                            int(math.ceil(l)),
                            len(each_range[a_slice])
                        )

                        a_slice = list()
                        a_slice.append(0 if start is None else start)
                        a_slice.append(0 if stop is None else stop)
                        a_slice.append(0 if step is None else step)

                        a_op = operator.itemgetter(*a_slice)

                        expected_result = None
                        try:
                            expected_result = a_op(each_range)
                        except IndexError:
                            pass

                        if expected_result is not None:
                            rf_slice = format.reformat_slice(a_slice)
                            rf_op = operator.itemgetter(*rf_slice)
                            self.assertEqual(
                                expected_result,
                                rf_op(each_range)
                            )

                            rf_slice = format.reformat_slice(a_slice, size)
                            rf_op = operator.itemgetter(*rf_slice)
                            self.assertEqual(
                                expected_result,
                                rf_op(each_range)
                            )
                        else:
                            format.reformat_slice(a_slice)
                            with self.assertRaises(IndexError):
                                format.reformat_slice(a_slice, size)

                if start is not None:
                    a_slice = start

                    expected_result = None
                    try:
                        expected_result = each_range[a_slice]
                    except IndexError:
                        pass

                    if expected_result is not None:
                        rf_slice = format.reformat_slice(a_slice)
                        self.assertEqual(
                            expected_result,
                            each_range[rf_slice]
                        )

                        rf_slice = format.reformat_slice(a_slice, size)
                        self.assertEqual(
                            expected_result,
                            each_range[rf_slice]
                        )
                    else:
                        format.reformat_slice(a_slice)
                        with self.assertRaises(IndexError):
                            format.reformat_slice(a_slice, size)

            rf_slice = format.reformat_slice(Ellipsis)
            self.assertEqual(
                each_range[:],
                each_range[rf_slice]
            )

            rf_slice = format.reformat_slice(Ellipsis, size)
            self.assertEqual(
                each_range[:],
                each_range[rf_slice]
            )

            rf_slice = format.reformat_slice(tuple())
            self.assertEqual(
                each_range[:],
                each_range[rf_slice]
            )

            rf_slice = format.reformat_slice(tuple(), size)
            self.assertEqual(
                each_range[:],
                each_range[rf_slice]
            )

            start = rf_slice.start
            stop = rf_slice.stop
            step = rf_slice.step

            if step is not None and step < 0 and stop is None:
                stop = -1

            l = float(stop - start)/float(step)
            self.assertEqual(
                int(math.ceil(l)),
                len(each_range[:])
            )


    def test_reformat_slices(self):
        with self.assertRaises(ValueError) as e:
            format.reformat_slices((slice(None),), (1, 2))

        self.assertEqual(
            str(e.exception),
            "Shape must be the same as the number of slices."
        )

        with self.assertRaises(ValueError) as e:
            format.reformat_slices(
                (slice(None), slice(None), Ellipsis), (1,)
            )

        self.assertEqual(
            str(e.exception),
            "Shape must be as large or larger than the number of slices"
            " without the Ellipsis."
        )

        with self.assertRaises(ValueError) as e:
            format.reformat_slices(
                (Ellipsis, Ellipsis), (1,)
            )

        self.assertEqual(
            str(e.exception),
            "Only one Ellipsis is permitted. Found multiple."
        )

        with self.assertRaises(ValueError) as e:
            format.reformat_slices(
                ([0, 1], [0, 1]),
            )

        self.assertEqual(
            str(e.exception),
            "Only one integral sequence supported. Instead got `2`."
        )

        rf_slice = format.reformat_slices(slice(None))
        self.assertEqual(
            rf_slice,
            (slice(0, None, 1),)
        )

        rf_slice = format.reformat_slices((slice(None),))
        self.assertEqual(
            rf_slice,
            (slice(0, None, 1),)
        )

        rf_slice = format.reformat_slices(Ellipsis)
        self.assertEqual(
            rf_slice,
            (Ellipsis,)
        )

        rf_slice = format.reformat_slices(Ellipsis, 10)
        self.assertEqual(
            rf_slice,
            (slice(0, 10, 1),)
        )

        rf_slice = format.reformat_slices(tuple())
        self.assertEqual(
            rf_slice,
            (Ellipsis,)
        )

        rf_slice = format.reformat_slices(tuple(), 10)
        self.assertEqual(
            rf_slice,
            (slice(0, 10, 1),)
        )

        rf_slice = format.reformat_slices(slice(None), 10)
        self.assertEqual(
            rf_slice,
            (slice(0, 10, 1),)
        )

        rf_slice = format.reformat_slices((slice(None),), 10)
        self.assertEqual(
            rf_slice,
            (slice(0, 10, 1),)
        )

        rf_slice = format.reformat_slices((
            -1,
            slice(None),
            slice(3, None),
            slice(None, 5),
            slice(None, None, 2)
        ))
        self.assertEqual(
            rf_slice,
            (
                -1,
                slice(0, None, 1),
                slice(3, None, 1),
                slice(0, 5, 1),
                slice(0, None, 2)
            )
        )

        rf_slice = format.reformat_slices(
            (
                -1,
                slice(None),
                slice(3, None),
                slice(None, 5),
                slice(None, None, 2),
                [-1, -2, -1, 1, 5]
            ),
            (12, 10, 13, 15, 20, 10)
        )
        self.assertEqual(
            rf_slice,
            (
                11,
                slice(0, 10, 1),
                slice(3, 13, 1),
                slice(0, 5, 1),
                slice(0, 20, 2),
                [9, 8, 9, 1, 5]
            )
        )

        rf_slice = format.reformat_slices(
            Ellipsis,
            (2, 3, 4, 5)
        )
        self.assertEqual(
            rf_slice,
            (
                slice(0, 2, 1),
                slice(0, 3, 1),
                slice(0, 4, 1),
                slice(0, 5, 1)
            )
        )

        rf_slice = format.reformat_slices(
            (
                Ellipsis,
                slice(0, 1)
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            rf_slice,
            (
                slice(0, 2, 1),
                slice(0, 3, 1),
                slice(0, 4, 1),
                slice(0, 1, 1)
            )
        )

        rf_slice = format.reformat_slices(
            (
                slice(0, 1),
                Ellipsis
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            rf_slice,
            (
                slice(0, 1, 1),
                slice(0, 3, 1),
                slice(0, 4, 1),
                slice(0, 5, 1)
            )
        )

        rf_slice = format.reformat_slices(
            (
                slice(0, 1),
                Ellipsis,
                slice(0, 1)
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            rf_slice,
            (
                slice(0, 1, 1),
                slice(0, 3, 1),
                slice(0, 4, 1),
                slice(0, 1, 1)
            )
        )

        rf_slice = format.reformat_slices(
            (
                slice(0, 1),
                Ellipsis,
                slice(0, 1),
                slice(0, 1),
                slice(0, 1)
            ),
            (2, 3, 4, 5)
        )
        self.assertEqual(
            rf_slice,
            (
                slice(0, 1, 1),
                slice(0, 1, 1),
                slice(0, 1, 1),
                slice(0, 1, 1)
            )
        )


    def test_split_indices(self):
        with self.assertRaises(ValueError) as e:
            format.split_indices(
                ([0, 1], [0, 1]),
            )

        self.assertEqual(
            str(e.exception),
            "Only one integral sequence supported. Instead got `2`."
        )

        sp_slice = format.split_indices(
            (3, Ellipsis, 0, slice(2, 5, 1), -1)
        )
        self.assertEqual(
            sp_slice,
            [
                (3, Ellipsis, 0, slice(2, 5, 1), -1)
            ]
        )

        sp_slice = format.split_indices(
            (3, Ellipsis, 0, slice(2, 5, 1), [-1])
        )
        self.assertEqual(
            sp_slice,
            [
                (3, Ellipsis, 0, slice(2, 5, 1), slice(-1, -2, -1))
            ]
        )

        sp_slice = format.split_indices(
            (3, Ellipsis, [0], slice(2, 5, 1), -1)
        )
        self.assertEqual(
            sp_slice,
            [
                (3, Ellipsis, slice(0, 1, 1), slice(2, 5, 1), -1)
            ]
        )

        sp_slice = format.split_indices(
            (3, Ellipsis, [0, 1, 2], slice(2, 5, 1), -1)
        )
        self.assertEqual(
            sp_slice,
            [
                (3, Ellipsis, slice(0, 1, 1), slice(2, 5, 1), -1),
                (3, Ellipsis, slice(1, 2, 1), slice(2, 5, 1), -1),
                (3, Ellipsis, slice(2, 3, 1), slice(2, 5, 1), -1)
            ]
        )

        sp_slice = format.split_indices(
            (3, Ellipsis, [2, 0, 1, 2], slice(2, 5, 1), -1)
        )
        self.assertEqual(
            sp_slice,
            [
                (3, Ellipsis, slice(2, 3, 1), slice(2, 5, 1), -1),
                (3, Ellipsis, slice(0, 1, 1), slice(2, 5, 1), -1),
                (3, Ellipsis, slice(1, 2, 1), slice(2, 5, 1), -1),
                (3, Ellipsis, slice(2, 3, 1), slice(2, 5, 1), -1)
            ]
        )



"""
Warning:
    The module ``kenjutsu.kenjutsu`` is deprecated. Please use
    ``kenjutsu.core`` instead.
"""

from __future__ import absolute_import


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Sep 08, 2016 15:46:46 EDT$"


import warnings


warnings.warn(
    "Please use `kenjutsu.core` instead.",
    DeprecationWarning
)


def reformat_slice(a_slice, a_length=None):
    """
        Takes a slice and reformats it to fill in as many undefined values as
        possible.

        Warning:
            This function is deprecated. Please use
            ``kenjutsu.core.reformat_slice`` instead.

        Args:
            a_slice(slice):        a slice to reformat.

            a_length(int):         a length to fill for stopping if not
                                   provided.

        Returns:
            (slice):               a new slice with as many values filled in as
                                   possible.

        Examples:

            >>> reformat_slice(slice(2, -1, None))
            slice(2, -1, 1)

            >>> reformat_slice(slice(2, -1, None), 10)
            slice(2, 9, 1)
    """

    warnings.warn(
        "Please use `kenjutsu.core.reformat_slice` instead.",
        DeprecationWarning
    )

    from kenjutsu import core

    return core.reformat_slice(a_slice, a_length)


def reformat_slices(slices, lengths=None):
    """
        Takes a tuple of slices and reformats them to fill in as many undefined
        values as possible.

        Warning:
            This function is deprecated. Please use
            ``kenjutsu.core.reformat_slices`` instead.

        Args:
            slices(tuple(slice)):        a tuple of slices to reformat.
            lengths(tuple(int)):         a tuple of lengths to fill.

        Returns:
            (slice):                     a tuple of slices with all default
                                         values filled if possible.

        Examples:

            >>> reformat_slices(
            ...     (
            ...         slice(None),
            ...         slice(3, None),
            ...         slice(None, 5),
            ...         slice(None, None, 2)
            ...     ),
            ...     (10, 13, 15, 20)
            ... )
            (slice(0, 10, 1), slice(3, 13, 1), slice(0, 5, 1), slice(0, 20, 2))
    """

    warnings.warn(
        "Please use `kenjutsu.core.reformat_slices` instead.",
        DeprecationWarning
    )

    from kenjutsu import core

    return core.reformat_slices(slices, lengths)


from kenjutsu.core import UnknownSliceLengthException


def len_slice(a_slice, a_length=None):
    """
        Determines how many elements a slice will contain.

        Warning:
            This function is deprecated. Please use
            ``kenjutsu.core.len_slice`` instead.

        Raises:
            UnknownSliceLengthException: Will raise an exception if
            a_slice.stop and a_length is None.

        Args:
            a_slice(slice):        a slice to reformat.

            a_length(int):         a length to fill for stopping if not
                                   provided.

        Returns:
            (slice):               a new slice with as many values filled in as
                                   possible.

        Examples:

            >>> len_slice(slice(2, None), 10)
            8

            >>> len_slice(slice(2, 6))
            4
    """

    warnings.warn(
        "Please use `kenjutsu.core.len_slice` instead.",
        DeprecationWarning
    )

    from kenjutsu import core

    return core.len_slice(a_slice, a_length)


def len_slices(slices, lengths=None):
    """
        Takes a tuple of slices and reformats them to fill in as many undefined
        values as possible.

        Warning:
            This function is deprecated. Please use
            ``kenjutsu.core.len_slices`` instead.

        Args:
            slices(tuple(slice)):        a tuple of slices to reformat.
            lengths(tuple(int)):         a tuple of lengths to fill.

        Returns:
            (slice):                     a tuple of slices with all default
                                         values filled if possible.

        Examples:

            >>> len_slices(
            ...     (
            ...         slice(None),
            ...         slice(3, None),
            ...         slice(None, 5),
            ...         slice(None, None, 2)
            ...     ),
            ...     (10, 13, 15, 20)
            ... )
            (10, 10, 5, 10)
    """

    warnings.warn(
        "Please use `kenjutsu.core.len_slices` instead.",
        DeprecationWarning
    )

    from kenjutsu import core

    return core.len_slices(slices, lengths)


def split_blocks(space_shape, block_shape, block_halo=None):
    """
        Return a list of slicings to cut each block out of an array or other.

        Takes an array with ``space_shape`` and ``block_shape`` for every
        dimension and a ``block_halo`` to extend each block on each side. From
        this, it can compute slicings to use for cutting each block out from
        the original array, HDF5 dataset or other.

        Warning:
            This function is deprecated. Please use
            ``kenjutsu.core.split_blocks`` instead.

        Note:
            Blocks on the boundary that cannot extend the full range will
            be truncated to the largest block that will fit. This will raise
            a warning, which can be converted to an exception, if needed.

        Args:
            space_shape(tuple):            Shape of array to slice
            block_shape(tuple):            Size of each block to take
            block_halo(tuple):             Halo to tack on to each block

        Returns:
            collections.Sequence of \
            tuples of slices:              Provides tuples of slices for \
                                           retrieving blocks.

        Examples:

            >>> split_blocks(
            ...     (2, 3,), (1, 1,), (1, 1,)
            ... )  #doctest: +NORMALIZE_WHITESPACE
            ([(slice(0, 1, 1), slice(0, 1, 1)),
              (slice(0, 1, 1), slice(1, 2, 1)),
              (slice(0, 1, 1), slice(2, 3, 1)),
              (slice(1, 2, 1), slice(0, 1, 1)),
              (slice(1, 2, 1), slice(1, 2, 1)),
              (slice(1, 2, 1), slice(2, 3, 1))],
            <BLANKLINE>
             [(slice(0, 2, 1), slice(0, 2, 1)),
              (slice(0, 2, 1), slice(0, 3, 1)),
              (slice(0, 2, 1), slice(1, 3, 1)),
              (slice(0, 2, 1), slice(0, 2, 1)),
              (slice(0, 2, 1), slice(0, 3, 1)),
              (slice(0, 2, 1), slice(1, 3, 1))],
            <BLANKLINE>
             [(slice(0, 1, 1), slice(0, 1, 1)),
              (slice(0, 1, 1), slice(1, 2, 1)),
              (slice(0, 1, 1), slice(1, 2, 1)),
              (slice(1, 2, 1), slice(0, 1, 1)),
              (slice(1, 2, 1), slice(1, 2, 1)),
              (slice(1, 2, 1), slice(1, 2, 1))])

    """

    warnings.warn(
        "Please use `kenjutsu.core.split_blocks` instead.",
        DeprecationWarning
    )

    from kenjutsu import core

    return core.split_blocks(space_shape, block_shape, block_halo)
